"""Exceptions raised by `pyserials.validate` module."""

from __future__ import annotations
from typing import Any as _Any

import jsonschema as _jsonschema
import mdit as _mdit
from protocolman import Stringable

from pyserials import write as _write
from pyserials.exception import _base


class PySerialsValidateException(_base.PySerialsException):
    """Base class for all exceptions raised by `pyserials.validate` module.

    Attributes
    ----------
    data : dict | list | str | int | float | bool
        The data that failed validation.
    schema : dict
        The schema that the data failed to validate against.
    validator : Any
        The validator that was used to validate the data against the schema.
    """

    def __init__(
        self,
        problem,
        data: dict | list | str | int | float | bool,
        schema: dict,
        validator: _Any,
        registry: _Any = None,
        section: dict | None = None,
    ):
        intro = _mdit.inline_container(
            "Failed to validate data against schema using validator ",
            _mdit.element.code_span(validator.__class__.__name__),
            "."
        )
        report = _mdit.document(
            heading="Schema Validation Error",
            body={
                "intro": intro,
                "problem": problem,
            },
            section=section,
        )
        super().__init__(report)
        self.data = data
        self.schema = schema
        self.validator = validator
        self.registry = registry
        return


class PySerialsInvalidJsonSchemaError(PySerialsValidateException):
    """Exception raised when data validation fails due to the schema being invalid."""

    def __init__(
        self,
        data: dict | list | str | int | float | bool,
        schema: dict,
        validator: _Any,
        registry: _Any = None,
    ):
        super().__init__(
            problem="The schema is invalid.",
            data=data,
            schema=schema,
            validator=validator,
            registry=registry,
        )
        return


class PySerialsJsonSchemaValidationError(PySerialsValidateException):
    """Exception raised when data validation fails due to the data being invalid against the schema."""

    def __init__(
        self,
        causes: list[_jsonschema.exceptions.ValidationError],
        data: dict | list | str | int | float | bool,
        schema: dict,
        validator: _Any,
        registry: _Any = None,
    ):
        self.causes = causes
        super().__init__(
            problem=self._generate_problem_statement(),
            section={str(idx): self._generate_error_report(error) for idx, error in enumerate(self.causes)},
            data=data,
            schema=schema,
            validator=validator,
            registry=registry,
        )
        return

    def _generate_problem_statement(self):
        error_paths = [_mdit.element.code_span(error.json_path) for error in self.causes]
        error_paths_str = self._join_list(error_paths)
        count_errors = len(error_paths)
        problem = _mdit.inline_container(
            "Found ",
            "an error " if count_errors == 1 else f"{count_errors} errors ",
            "in the data at ",
            error_paths_str,
            "."
        )
        return problem

    def _generate_error_report(
        self,
        error: _jsonschema.exceptions.ValidationError,
    ) -> _mdit.Document:
        problem = self._parse_error_message(error)
        short_ver_fieldlist_items = [
            _mdit.element.field_list_item("Problem", problem),
            _mdit.element.field_list_item(
                "Validator Path", _mdit.element.code_span(self._create_path(error.absolute_schema_path))
            ),
        ]
        full_ver_items = [
            _mdit.inline_container(problem),
            self._make_yaml_code_admo(
                admo_type="error",
                title="Validator",
                title_details=str(error.validator),
                content=error.validator_value,
            ),
            self._make_yaml_code_admo(
                admo_type="error",
                title="Schema",
                title_details=self._create_path(error.absolute_schema_path),
                content=error.schema,
            ),
            self._make_yaml_code_admo(
                admo_type="error",
                title="Instance",
                title_details=self._create_path(error.absolute_path),
                content=error.instance,
            ),
        ]
        section = {}
        if error.context:
            context_paths = []
            for idx, sub_error in enumerate(sorted(error.context, key=lambda x: len(x.context))):
                section[str(idx)] = self._generate_error_report(sub_error)
                context_paths.append(_mdit.element.code_span(sub_error.json_path))
            context_paths_joined = self._join_list(context_paths)
            short_ver_fieldlist_items.insert(
                1,
                _mdit.element.field_list_item(
                    f"Context Path{'s' if len(context_paths) > 1 else ''}",
                    context_paths_joined
                )
            )
            err = "an error" if len(context_paths) == 1 else f"{len(context_paths)} errors"
            full_ver_items[0].append(
                f" This was caused by {err} at {context_paths_joined}.",
            )
        doc = _mdit.document(
            heading=_mdit.element.code_span(error.json_path),
            body={
                "short": (_mdit.element.field_list(short_ver_fieldlist_items), ("short", "console")),
                "full": (_mdit.block_container(*full_ver_items), "full"),
            },
            section=section,
        )
        return doc

    @staticmethod
    def _make_yaml_code_admo(admo_type: str, title: str, title_details: str, content: dict) -> _mdit.element.Admonition:
        code_block = _mdit.element.code_block(
            content=_write.to_yaml_string(content, end_of_file_newline=False),
            language="yaml",
        )
        admo = _mdit.element.admonition(
            title=_mdit.inline_container(f"**{title}**: ", _mdit.element.code_span(title_details)),
            body=code_block,
            type=admo_type,
            dropdown=True,
        )
        return admo

    @staticmethod
    def _parse_error_message(error: _jsonschema.exceptions.ValidationError) -> str:
        instance_str = str(error.instance)
        if error.message.startswith(instance_str):
            msg = error.message.removeprefix(str(error.instance)).strip()
            problem = f"Data {msg}"
        else:
            problem = error.message
        return f"{problem.removeprefix(".")}."

    @staticmethod
    def _create_path(path):
        return "$." + ".".join(str(path_component) for path_component in path)

    @staticmethod
    def _join_list(
        items: list,
        sep: Stringable = ", ",
        sep_last: Stringable = ", and ",
        sep_pair: Stringable = " and ",
    ) -> Stringable:
        if len(items) == 1:
            return items[0]
        elif len(items) == 2:
            return _mdit.inline_container(items[0], sep_pair, items[1])
        container = []
        for item in items[:-1]:
            container.extend([item, sep])
        container.pop()
        container.extend([sep_last, items[-1]])
        return _mdit.inline_container(*container)


"""Exceptions raised by `pyserials.read` module."""

from __future__ import annotations
from typing import Literal as _Literal
from pathlib import Path as _Path

import ruamel.yaml as _yaml
import json as _json
import mdit as _mdit

from tomlkit.exceptions import TOMLKitError as _TOMLKitError

from pyserials.exception import _base


class PySerialsReadException(_base.PySerialsException):
    """Base class for all exceptions raised by `pyserials.read` module.

    Attributes
    ----------
    source_type : {"file", "string"}
        Type of source from which data was read.
    data_type : {"json", "yaml", "toml"} or None
        Type of input data, if known.
    filepath : pathlib.Path or None
        Path to the input datafile, if data was read from a file.
    """

    def __init__(
        self,
        source_type: _Literal["file", "string"],
        problem,
        section: dict | None = None,
        data_type: _Literal["json", "yaml", "toml"] | None = None,
        filepath: _Path | None = None,
    ):
        intro = _mdit.inline_container(
            "Failed to read",
            f"{data_type.upper()} data" if data_type else "data",
            "from input",
            "string." if source_type == "string" else _mdit.inline_container(
                "file at ", _mdit.element.code_span(filepath), "."
            ),
            separator=" ",
        )
        report = _mdit.document(
            heading="Data Read Error",
            body={
                "intro": intro,
                "problem": problem,
            },
            section=section,
        )
        super().__init__(report)
        self.source_type: _Literal["file", "string"] = source_type
        self.data_type: _Literal["json", "yaml", "toml"] | None = data_type
        self.filepath: _Path | None = filepath
        return


class PySerialsEmptyStringError(PySerialsReadException):
    """Exception raised when a string to be read is empty."""

    def __init__(self, data_type: _Literal["json", "yaml", "toml"]):
        problem = f"The string is empty."
        super().__init__(problem=problem, source_type="string", data_type=data_type)
        return


class PySerialsInvalidFileExtensionError(PySerialsReadException):
    """Exception raised when a file to be read has an unrecognized extension."""

    def __init__(self, filepath: _Path):
        problem = _mdit.inline_container(
            "The file extension must be one of ",
            _mdit.element.code_span('json'),
            ", ",
            _mdit.element.code_span('yaml'),
            ", ",
            _mdit.element.code_span('yml'),
            ", or ",
            _mdit.element.code_span('.toml'),
            ", but got ",
            _mdit.element.code_span(str(filepath.suffix.removeprefix('.'))),
            ". Please provide the extension explicitly, or rename the file to have a valid extension."
        )
        super().__init__(problem=problem, source_type="file", filepath=filepath)
        return


class PySerialsMissingFileError(PySerialsReadException):
    """Exception raised when a file to be read does not exist."""

    def __init__(self, data_type: _Literal["json", "yaml", "toml"], filepath: _Path):
        problem = f"The file does not exist."
        super().__init__(problem=problem, source_type="file", data_type=data_type, filepath=filepath)
        return


class PySerialsEmptyFileError(PySerialsReadException):
    """Exception raised when a file to be read is empty."""

    def __init__(self, data_type: _Literal["json", "yaml", "toml"], filepath: _Path):
        problem = f"The file is empty."
        super().__init__(problem=problem, source_type="file", data_type=data_type, filepath=filepath)
        return


class PySerialsInvalidDataError(PySerialsReadException):
    """Exception raised when the data is invalid.

    Attributes
    ----------
    data : str
        The input data that was supposed to be read.
    """

    def __init__(
        self,
        source_type: _Literal["file", "string"],
        data_type: _Literal["json", "yaml", "toml"],
        data: str,
        cause: Exception,
        filepath: _Path | None = None,
    ):
        self.data = data
        self.cause = cause
        self.problem: str = str(cause)
        self.problem_line: int | None = None
        self.problem_column: int | None = None
        self.problem_data_type: str | None = None
        self.context: str | None = None
        self.context_line: int | None = None
        self.context_column: int | None = None
        self.context_data_type: str | None = None
        self.data_type = data_type

        if isinstance(cause, _yaml.YAMLError):
            self.problem_line = cause.problem_mark.line + 1
            self.problem_column = cause.problem_mark.column + 1
            self.problem_data_type = cause.problem_mark.name.removeprefix("<").removesuffix(">")
            self.problem = cause.problem.strip()
            if cause.context:
                self.context = cause.context.strip()
                if cause.context_mark:
                    self.context_line = cause.context_mark.line + 1
                    self.context_column = cause.context_mark.column + 1
                    self.context_data_type = cause.context_mark.name.removeprefix("<").removesuffix(">")
        elif isinstance(cause, _json.JSONDecodeError):
            self.problem = cause.msg
            self.problem_line = cause.lineno
            self.problem_column = cause.colno
        elif isinstance(cause, _TOMLKitError):
            self.problem_line = cause.line
            self.problem_column = cause.col
            self.problem = cause.args[0].removesuffix(f" at line {self.problem_line} col {self.problem_column}")
        self.problem = self.problem.strip().capitalize().removesuffix(".")
        description = ["The data is not valid"]
        if self.problem_line:
            description.extend(
                [
                    " at line ",
                    _mdit.element.code_span(str(self.problem_line)),
                    ", column ",
                    _mdit.element.code_span(str(self.problem_column)),
                ]
            )
        description.append(f": {self.problem}.")
        super().__init__(
            problem=description,
            source_type=source_type,
            section=self._report_content(),
            data_type=data_type,
            filepath=filepath
        )
        return

    def _report_content(self) -> dict:

        def make_table(problem, line, column, data_type):
            items = [
                _mdit.element.field_list_item(title=title, body=value) for title, value in [
                    ["Description", problem],
                    ["Line Number", line],
                    ["Column Number", column],
                    ["Data Type", data_type],
                ] if value is not None
            ]
            return _mdit.element.field_list(items)

        content = {
            "problem_details": _mdit.element.admonition(
                title="Problem",
                body=make_table(self.problem, self.problem_line, self.problem_column, self.problem_data_type),
                type="error",
            )
        }
        if self.context:
            content["context_details"] = _mdit.element.admonition(
                title="Context",
                body=make_table(self.context, self.context_line, self.context_column, self.context_data_type),
                type="note",
            )

        code_block_full = _mdit.element.code_block(
            content=self.data,
            language=self.data_type,
            caption="Data",
            line_num=True,
            emphasize_lines=[line for line in (self.problem_line, self.context_line) if line],
            degrade_to_diff=True,
        )
        content["data_full"] = (code_block_full, "full")

        if not (self.problem_line or self.context_line):
            code_block_short = _mdit.element.code_block(
                content=self.data[:1000].strip() + "\n..." if len(self.data) > 1000 else self.data,
                language=self.data_type,
                caption="Data" if len(self.data) <= 1000 else "Data (truncated to first 1000 characters)",
                line_num=True,
            )
        else:
            if self.problem_line and self.context_line:
                line_start = min(self.problem_line, self.context_line)
                line_end = max(self.problem_line, self.context_line)
            else:
                line_start = self.problem_line or self.context_line
                line_end = line_start
            data_lines = self.data.splitlines()
            selected_lines = data_lines[line_start - 1:line_end]
            code_block_short = _mdit.element.code_block(
                content="\n".join(selected_lines),
                language=self.data_type,
                caption="Data",
                line_num=True,
                line_num_start=line_start,
                emphasize_lines=list({1, len(selected_lines)}),
            )
        content["data_short"] = (code_block_short, ("short", "console"))
        container = _mdit.block_container(content)
        doc = _mdit.document(heading="Error Details", body=container)
        return {"details": doc}



"""Exceptions raised by `pyserials.update` module."""

from __future__ import annotations
from typing import Any as _Any, Literal as _Literal

import mdit as _mdit

from pyserials.exception import _base


class PySerialsUpdateException(_base.PySerialsException):
    """Base class for all exceptions raised by `pyserials.update` module.

    Attributes
    ----------
    path : str
        JSONPath to where the update failed.
    data : dict | list | str | int | float | bool
        Data that failed to update.
    data_full : dict | list | str | int | float | bool
        Full data input.
    """

    def __init__(
        self,
        path: str,
        data: dict | list | str | int | float | bool,
        data_full: dict | list | str | int | float | bool,
        problem,
        section: dict | None = None,
    ):
        intro = _mdit.inline_container(
            "Failed to update data at ",
            _mdit.element.code_span(path),
            "."
        )
        report = _mdit.document(
            heading="Data Update Error",
            body={
                "intro": intro,
                "problem": problem,
            },
            section=section,
        )
        super().__init__(report)
        self.path = path
        self.data = data
        self.data_full = data_full
        return


class PySerialsUpdateRecursiveDataError(PySerialsUpdateException):
    """Base class for all exceptions raised by `pyserials.update.recursive_update`.

    Attributes
    ----------
    data_addon : Any
        Value of the failed data in the addon dictionary.
    data_addon_full : dictionary
        Full addon input.
    """

    def __init__(
        self,
        problem_type: _Literal["duplicate", "type_mismatch"],
        path: str,
        data: _Any,
        data_full: dict,
        data_addon: _Any,
        data_addon_full: dict,
    ):
        self.type_data = type(data)
        self.type_data_addon = type(data_addon)
        problem = _mdit.inline_container(
            "There was a duplicate in the addon dictionary: ",
            "the value of type",
            _mdit.element.code_span(self.type_data_addon.__name__),
            " already exists in the source data."
        ) if problem_type == "duplicate" else _mdit.inline_container(
            "There was a type mismatch between the source and addon dictionary values: ",
            "the value is of type ",
            _mdit.element.code_span(self.type_data.__name__),
            " in the source data, ",
            "but of type ",
            _mdit.element.code_span(self.type_data_addon.__name__),
            " in the addon data."
        )
        super().__init__(
            path=path,
            data=data,
            data_full=data_full,
            problem=problem,
        )
        self.problem_type: _Literal["duplicate", "type_mismatch"] = problem_type
        self.data_addon = data_addon
        self.data_addon_full = data_addon_full
        return


class PySerialsUpdateTemplatedDataError(PySerialsUpdateException):
    """Exception raised when updating templated data fails.

    Attributes
    ----------
    path_invalid : str
        JSONPath that caused the update to fail.
    data_source : dict
        Source data that was used to update the template.
    template_start : str
        The start marker of the template.
    template_end : str
        The end marker of the template.
    """

    def __init__(
        self,
        description_template: str,
        path_invalid: str,
        path: str,
        data: str,
        data_full: dict | list | str | int | float | bool,
        data_source: dict,
        template_start: str,
        template_end: str,
    ):
        self.path_invalid = path_invalid
        self.data_source = data_source
        self.template_start = template_start
        self.template_end = template_end
        parts = description_template.split("{path_invalid}")
        if len(parts) > 1:
            parts.insert(1, _mdit.element.code_span(str(self.path_invalid)))
        super().__init__(
            path=str(path),
            data=data,
            data_full=data_full,
            problem=_mdit.inline_container(*parts),
        )
        return



from __future__ import annotations as _annotations

from typing import TYPE_CHECKING as _TYPE_CHECKING

import pyserials as _ps

if _TYPE_CHECKING:
    from typing import Callable, Any


class NestedDict:

    def __init__(
        self,
        data: dict | None = None,
        marker_start_value: str = "$",
        marker_end_value: str = "$",
        repeater_start_value: str = "{",
        repeater_end_value: str = "}",
        repeater_count_value: int = 2,
        start_list: str = "$[[",
        start_unpack: str = "*{{",
        start_code: str = "#{{",
        end_list: str = "]]$",
        end_unpack: str = "}}*",
        end_code: str = "}}#",
        raise_no_match: bool = True,
        leave_no_match: bool = False,
        no_match_value: Any = None,
        code_context: dict[str, Any] | None = None,
        code_context_partial: dict[str, Callable | tuple[Callable, str]] | None = None,
        code_context_call: dict[str, Callable[[Callable], Any]] | None = None,
        stringer: Callable[[str], str] = str,
        unpack_string_joiner: str = ", ",
        relative_template_keys: list[str] | None = None,
        relative_key_key: str | None = None,
        implicit_root: bool = True,
        getter_function_name: str = "get",
        skip_key_func: Callable[[list[str]], bool] | None = None,
    ):
        self._data = data or {}
        self._templater = _ps.update.TemplateFiller(
            marker_start_value=marker_start_value,
            marker_end_value=marker_end_value,
            repeater_start_value=repeater_start_value,
            repeater_end_value=repeater_end_value,
            repeater_count_value=repeater_count_value,
            start_list=start_list,
            start_unpack=start_unpack,
            start_code=start_code,
            end_list=end_list,
            end_unpack=end_unpack,
            end_code=end_code,
            raise_no_match=raise_no_match,
            leave_no_match=leave_no_match,
            no_match_value=no_match_value,
            code_context=code_context,
            code_context_partial=code_context_partial,
            code_context_call=code_context_call,
            stringer=stringer,
            unpack_string_joiner=unpack_string_joiner,
            relative_template_keys=relative_template_keys,
            relative_key_key=relative_key_key,
            implicit_root=implicit_root,
            getter_function_name=getter_function_name,
            skip_key_func=skip_key_func,
        )
        return

    def fill(self, path: str = ""):
        if not path:
            value = self._data
        else:
            value = self.__getitem__(path)
        if not value:
            return
        filled_value = self.fill_data(
            data=value,
            current_path=path,
        )
        if not path:
            self._data = filled_value
        else:
            self.__setitem__(path, filled_value)
        return filled_value

    def fill_data(self, data, current_path: str = ""):
        return self._templater.fill(
            data=self._data,
            template=data,
            current_path=current_path,
        )

    def __call__(self):
        return self._data

    def __getitem__(self, item: str):
        keys = item.split(".")
        data = self._data
        for key in keys:
            if not isinstance(data, dict):
                raise KeyError(f"Key '{key}' not found in '{data}'.")
            if key not in data:
                return
            data = data[key]
        # if isinstance(data, dict):
        #     return NestedDict(data)
        # if isinstance(data, list) and all(isinstance(item, dict) for item in data):
        #     return [NestedDict(item) for item in data]
        return data

    def __setitem__(self, key, value):
        key = key.split(".")
        data = self._data
        for k in key[:-1]:
            if k not in data:
                data[k] = {}
            data = data[k]
        data[key[-1]] = value
        return

    def __contains__(self, item):
        keys = item.split(".")
        data = self._data
        for key in keys:
            if not isinstance(data, dict) or key not in data:
                return False
            data = data[key]
        return True

    def __bool__(self):
        return bool(self._data)

    def setdefault(self, key, value):
        key = key.split(".")
        data = self._data
        for k in key[:-1]:
            if k not in data:
                data[k] = {}
            data = data[k]
        return data.setdefault(key[-1], value)

    def get(self, key, default=None):
        keys = key.split(".")
        data = self._data
        for key in keys:
            if not isinstance(data, dict) or key not in data:
                return default
            data = data[key]
        return data

    def items(self):
        return self._data.items()

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()

    def update(self, data: dict):
        self._data.update(data)
        return

    def pop(self, key: str, default=None):
        keys = key.split(".")
        data = self._data
        for key in keys[:-1]:
            if not isinstance(data, dict) or key not in data:
                return default
            data = data[key]
        if not isinstance(data, dict) or key not in data:
            return default
        return data.pop(key)



"""PySerials"""

from pyserials import exception, update, validate, read, write, format, compare
from pyserials.nested_dict import NestedDict
from pyserials.property_dict import PropertyDict
from pyserials.flatten import flatten



def flatten(data: dict | list[dict]) -> dict | list[dict]:
    """Flatten a nested dictionary into a single-level dictionary.

    Parameters
    ----------
    data
        A dictionary or a list of dictionaries to flatten.

    Returns
    -------
    If `data` is a dictionary, returns a flattened dictionary.
    If `data` is a list of dictionaries, returns a list of flattened dictionaries.
    """
    if isinstance(data, dict):
        return _flatten(data)
    if isinstance(data, list):
        if not data:
            return []
        if not all(isinstance(item, dict) for item in data):
            raise ValueError("All items in the list must be dictionaries.")
        return [_flatten(item) for item in data]
    raise ValueError("Input must be a dictionary or a list of dictionaries.")


def _flatten(data, output: dict | None = None, path: str | None = None):
    output = output if output is not None else {}
    if isinstance(data, str | int | float | bool | None):
        output[path] = data
        return output
    if isinstance(data, dict):
        for key, value in data.items():
            new_path = _flatten_key(path, key)
            _flatten(data=value, output=output, path=new_path)
        return output
    if isinstance(data, list):
        if len(data) == 0:
            output[path] = None
            return output
        if all(isinstance(val, str | int | float | bool | None) for val in data):
            output[path] = data[0] if len(data) == 1 else data
            return output
        if len(data) != 1:
            raise ValueError(f"Data at {path} is a list with more than one non-primitive element: {data}")
        return _flatten(data=data[0], output=output, path=path)
    raise ValueError(f"Unknown data type {type(data)} at {path}")


def _flatten_key(path: str | None, key: str):
    if path:
        return f"{path}.{key}"
    return key



import tomlkit as _tomlkit
import ruamel.yaml as _yaml
import ruamel.yaml.comments as _yaml_comments
from typing import Literal as _Literal


def to_toml_object(
    data: str | list | dict,
    toml_type: _Literal[
        "str", "table", "array", "inline_table", "array_of_inline_tables", "table_of_arrays", "table_of_tables"
    ],
):
    match toml_type:
        case "str":
            return data
        case "table":
            return data
        case "array":
            array = _tomlkit.array().multiline(True)
            array.extend(data)
            return array
        case "inline_table":
            inline = _tomlkit.inline_table()
            inline.update(data)
            return inline
        case "array_of_inline_tables":
            arr = _tomlkit.array().multiline(True)
            for table in data:
                tab = _tomlkit.inline_table()
                tab.update(table)
                arr.append(tab)
            return arr
        case "table_of_arrays":
            return {tab_key: _tomlkit.array(arr).multiline(True) for tab_key, arr in data.items()}
        case "table_of_tables":
            return _tomlkit.table(is_super_table=True).update(data)
        case _:
            raise ValueError(f"Unknown data type {toml_type}.")


def to_yaml_array(
    data: list,
    inline: bool = True,
) -> _yaml_comments.CommentedSeq:
    """Convert a list to a YAML array.

    Parameters
    ----------
    data : list
        The list to convert.
    inline : bool
        Whether to use inline (i.e., flow) or block style.
    """
    array = _yaml_comments.CommentedSeq()
    array.fa.set_flow_style() if inline else array.fa.set_block_style()
    array.extend(data)
    return array
